from collections import defaultdict
from itertools import product

import numpy as np

from ms_peak_picker import FittedPeak

from .feature_map import (
    LCMSFeatureMap,
    DeconvolutedLCMSFeatureMap)
from .lcms_feature import (
    LCMSFeature,
    EmptyFeature,
    FeatureSetIterator)
from .feature_fit import (
    LCMSFeatureSetFit,
    DeconvolutedLCMSFeatureTreeNode,
    DeconvolutedLCMSFeature)
from .dependence_network import FeatureDependenceGraph
from .profile_transform import binsearch, smooth_leveled
from ms_deisotope.peak_dependency_network.intervals import Interval, IntervalTreeNode
from ms_deisotope.averagine import (
    AveragineCache, PROTON, isotopic_shift,
    neutral_mass)
from ms_deisotope.peak_set import DeconvolutedPeak
from ms_deisotope.envelope_statistics import (
    most_abundant_mz, average_mz, a_to_a2_ratio)
from ms_deisotope.deconvolution import (
    charge_range_, drop_placeholders, first_peak,
    mean)
from ms_deisotope.utils import printer


def conform_envelopes(experimental, base_theoretical, minimum_theoretical_abundance=0.05):
    total = 0
    n_missing = 0
    i = 0
    cleaned_eid = []
    for peak in experimental:
        if peak is None:
            peak = FittedPeak(base_theoretical[i].mz, 1, 1, -1, -1, 0, 1, 0, 0)
            if base_theoretical[i].intensity > minimum_theoretical_abundance:
                n_missing += 1
        total += peak.intensity
        cleaned_eid.append(peak)
        i += 1

    tid = []
    for peak in base_theoretical:
        peak = peak.clone()
        peak.intensity *= total
        tid.append(peak)
    return cleaned_eid, tid, n_missing


class LCMSFeatureProcessorBase(object):
    def create_theoretical_distribution(self, mz, charge, charge_carrier=PROTON, truncate_after=0.8,
                                        ignore_below=0.05):
        base_tid = self.averagine.isotopic_cluster(
            mz, charge, truncate_after=truncate_after, charge_carrier=charge_carrier)
        total = 0
        kept_tid = []
        for i, p in enumerate(base_tid):
            if p.intensity < ignore_below and i > 1:
                continue
            else:
                total += p.intensity
                kept_tid.append(p)
        for p in kept_tid:
            p.intensity /= total
        return kept_tid

    def find_all_features(self, mz, error_tolerance=2e-5):
        return self.feature_map.find_all(mz, error_tolerance)

    def find_features(self, mz, error_tolerance=2e-5, interval=None):
        f = self.find_all_features(mz, error_tolerance)
        if not f:
            f = [None]
        elif interval is not None:
            f = [el for el in f if el.overlaps_in_time(interval)]
        return f

    def conform_envelopes(self, experimental, base_theoretical):
        return conform_envelopes(experimental, base_theoretical)

    def _find_thresholded_score(self, scores, percentage):
        scores = np.array(scores)
        maximum = scores.max()
        threshold = maximum * percentage
        return scores[scores > threshold].mean()

    def _fit_feature_set(self, mz, error_tolerance, charge,
                         charge_carrier=PROTON, truncate_after=0.8, max_missed_peaks=1,
                         threshold_scale=0.3, feature=None):
        base_tid = self.create_theoretical_distribution(mz, charge, charge_carrier, truncate_after)
        feature_groups = self.match_theoretical_isotopic_distribution(base_tid, error_tolerance, interval=feature)
        feature_fits = []
        for features in product(*feature_groups):
            if all(f is None for f in features):
                continue
            # If the monoisotopic feature wasn't actually observed, create a dummy feature
            # since the monoisotopic feature cannot be None
            if features[0] is None:
                features = list(features)
                features[0] = EmptyFeature(mz)
            feat_iter = FeatureSetIterator(features)
            scores = []
            times = []
            counter = 0
            for eid in feat_iter:
                cleaned_eid, tid, n_missing = self.conform_envelopes(eid, base_tid)
                if n_missing > max_missed_peaks:
                    continue
                score = self.scorer.evaluate(None, cleaned_eid, tid)
                if np.isnan(score):
                    continue
                scores.append(score)
                times.append(feat_iter.current_time)
                counter += 1
            final_score = self._find_thresholded_score(scores, threshold_scale)
            missing_features = 0
            for f in features:
                if f is None:
                    missing_features += 1
            fit = LCMSFeatureSetFit(
                features, base_tid, final_score, charge, missing_features,
                neutral_mass=neutral_mass(features[0].mz, charge, charge_carrier),
                missing_features=missing_features,
                scores=scores, times=times)
            if self.scorer.reject_score(fit.score):
                continue
            feature_fits.append(fit)
        return feature_fits

    def match_theoretical_isotopic_distribution(self, theoretical_distribution, error_tolerance=2e-5, interval=None):
        """Given a list of theoretical peaks, find their counterparts in :attr:`peaklist` within `error_tolerance`
        ppm error. If no experimental peak is found, a placeholder will be used in its stead.

        Parameters
        ----------
        theoretical_distribution : list of TheoreticalPeak
            The theoretical isotopic pattern to match
        error_tolerance : float, optional
            Parts-per-million error tolerance to permit in searching for matches

        Returns
        -------
        list of list of LCMSFeature
            The list of matched features for each theoretical m/z
        """
        experimental_distribution = [self.find_features(
            p.mz, error_tolerance, interval=interval) for p in theoretical_distribution]
        return experimental_distribution


try:
    has_c = True
    from ms_deisotope._c.feature_map.feature_processor import LCMSFeatureProcessorBase
except ImportError:
    has_c = False


def test_fit_features(self, features, charge, charge_carrier=PROTON, truncate_after=0.8, max_missed_peaks=1,
                      threshold_scale=0.3):
    base_tid = self.create_theoretical_distribution(features[0].mz, charge, charge_carrier, truncate_after)
    feat_iter = FeatureSetIterator(features)
    scores = []
    times = []
    counter = 0
    for eid in feat_iter:
        cleaned_eid, tid, n_missing = self.conform_envelopes(eid, base_tid)
        if n_missing > max_missed_peaks:
            continue
        score = self.scorer.evaluate(None, cleaned_eid, tid)
        if np.isnan(score):
            continue
        scores.append(score)
        times.append(feat_iter.current_time)
        counter += 1

    return np.array(scores), np.array(times)


def count_placeholders(peaks):
    """Counts the number of placeholder peaks in an iterable
    of FittedPeaks

    Parameters
    ----------
    peaks : Iterable of FittedPeak

    Returns
    -------
    int
        Number of placeholder peaks
    """
    i = 0
    for peak in peaks:
        if peak.intensity <= 1:
            i += 1
    return i


def test_finalize(self, feature_fit, charge_carrier=PROTON, detection_threshold=0.1,
                  max_missed_peaks=1):
    start_time, end_time = find_bounds(feature_fit, detection_threshold)
    feat_iter = FeatureSetIterator(
        feature_fit.features, start_time, end_time)
    base_tid = feature_fit.theoretical
    charge = feature_fit.charge
    abs_charge = abs(charge)
    for eid in feat_iter:
        cleaned_eid, tid, n_missing = conform_envelopes(eid, base_tid)
        rep_eid = drop_placeholders(cleaned_eid)
        n_real_peaks = len(rep_eid)
        invalid = n_real_peaks == 0 or (n_real_peaks == 1 and abs_charge > 1) or n_missing > max_missed_peaks
        score = self.scorer.evaluate(None, cleaned_eid, tid)
        yield feat_iter.current_time, score, n_missing, invalid


class PrecursorMap(object):

    @classmethod
    def extract_ms1_scans_with_msn(cls, peak_loader):
        mapping = defaultdict(list)
        for msn_id, msn_info in peak_loader.extended_index.msn_ids.items():
            prec_id = msn_info['precursor_scan_id']
            mapping[prec_id].append(msn_info)
        return mapping

    @classmethod
    def map_precursor_information_to_peaks(cls, peak_loader, error_tolerance=2e-5):
        index = cls.extract_ms1_scans_with_msn(peak_loader)
        mapping = dict()
        missed = []
        for precursor_id, msn_info_set in index.items():
            precursor = peak_loader.get_scan_by_id(precursor_id)
            for msn_info in msn_info_set:
                peak = precursor.peak_set.has_peak(msn_info['mz'], error_tolerance)
                if peak is not None:
                    msn_info['intensity'] = peak.intensity
                    mapping[precursor.scan_time, peak.index] = msn_info
                else:
                    missed.append(msn_info)
        return mapping, missed

    @classmethod
    def build(cls, peak_loader, error_tolerance=2e-5):
        mapping, missed = cls.map_precursor_information_to_peaks(
            peak_loader, error_tolerance)
        return cls(mapping, missed)

    def __init__(self, mapping, missed=None):
        if missed is None:
            missed = []
        self.mapping = dict(mapping)
        self.missed = missed

    def assign_precursors(self, feature):
        hits = []
        for node in feature:
            for peak in node.members:
                key = (node.time, peak.index)
                if key in self.mapping:
                    pinfo = self.mapping.pop(key)
                    hits.append((node, pinfo))
        return hits

    def precursor_for_peak(self, time_peak_pair):
        if time_peak_pair in self.mapping:
            pinfo = self.mapping.pop(time_peak_pair)
        else:
            pinfo = None
        return pinfo


class LCMSFeatureProcessor(LCMSFeatureProcessorBase):
    def __init__(self, feature_map, averagine, scorer, precursor_map=None, minimum_size=3,
                 maximum_time_gap=0.25):
        if precursor_map is None:
            precursor_map = PrecursorMap({})
        self.feature_map = LCMSFeatureMap([f.clone(deep=True) for f in feature_map])
        self.averagine = AveragineCache(averagine)
        self.scorer = scorer
        self.precursor_map = precursor_map
        self.minimum_size = minimum_size
        self.maximum_time_gap = maximum_time_gap
        self.dependence_network = FeatureDependenceGraph(self.feature_map)
        self.orphaned_nodes = []

    def select_best_disjoint_subgraphs(self, disjoint_envelopes):
        solutions = []
        for cluster in disjoint_envelopes:
            disjoint_best_fits = cluster.disjoint_best_fits()
            for fit in disjoint_best_fits:
                solutions.append(fit)
        return solutions

    def charge_state_determination(self, feature, error_tolerance, charge_range=(1, 8),
                                   left_search=1, right_search=1, charge_carrier=PROTON,
                                   truncate_after=0.8, max_missed_peaks=1, threshold_scale=0.3):
        fits = self.collect_all_fits(
            feature,
            error_tolerance=error_tolerance,
            charge_range=charge_range,
            left_search=left_search,
            right_search=right_search,
            charge_carrier=charge_carrier,
            truncate_after=truncate_after,
            max_missed_peaks=max_missed_peaks,
            threshold_scale=threshold_scale)
        return sorted(fits, key=lambda x: x.score, reverse=True)

    def collect_all_fits(self, feature, error_tolerance, charge_range=(1, 8),
                         left_search=1, right_search=1, charge_carrier=PROTON,
                         truncate_after=0.8, max_missed_peaks=1, threshold_scale=0.3):
        fits = []
        holdout = None
        if self.scorer.is_maximizing():
            best_fit_score = 0
        else:
            best_fit_score = float('inf')
        best_fit_charge = 0
        for charge in charge_range_(*charge_range):
            current_fits = self.fit_theoretical_distribution(
                feature,
                error_tolerance=error_tolerance,
                charge=charge,
                left_search=left_search,
                right_search=right_search,
                charge_carrier=charge_carrier,
                truncate_after=truncate_after,
                max_missed_peaks=max_missed_peaks,
                threshold_scale=threshold_scale)
            if self.scorer.is_maximizing():
                for fit in current_fits:
                    if fit.score > best_fit_score:
                        best_fit_score = fit.score
                        best_fit_charge = charge
            else:
                for fit in current_fits:
                    if fit.score < best_fit_score:
                        best_fit_score = fit.score
                        best_fit_charge = charge
            if abs(charge) == 1:
                holdout = current_fits
            else:
                for fit in current_fits:
                    self.dependence_network.add_fit_dependence(fit)
                    fits.append(fit)
        if holdout is not None and best_fit_charge == 1:
            for fit in holdout:
                self.dependence_network.add_fit_dependence(fit)
                fits.append(fit)
        return fits

    def fit_theoretical_distribution(self, feature, error_tolerance, charge, left_search=1, right_search=1,
                                     charge_carrier=PROTON, truncate_after=0.8, max_missed_peaks=1,
                                     threshold_scale=0.3):
        """Fits all theoretical distributions for `feature` with charge state `charge`.

        Parameters
        ----------
        feature : LCMSFeature
            The feature to use as the seed fit
        error_tolerance : float
            Permitted parts-per-million mass accuracy
        charge : int
            The charge state to fit at
        left_search : int, optional
            Description
        right_search : int, optional
            Description
        charge_carrier : float, optional
            Mass of the charge carrier
        truncate_after : float, optional
            Fraction of the total isotopic pattern to include
            in the set of theoretical peaks to search for.

        Returns
        -------
        list of LCMSFeatureSetFit
        """
        base_mz = feature.mz
        all_fits = []
        for offset in range(-left_search, right_search + 1):
            shift = isotopic_shift(charge) * offset
            mz = base_mz + shift
            feature_fits = self._fit_feature_set(
                mz,
                error_tolerance=error_tolerance,
                charge=charge,
                charge_carrier=charge_carrier,
                truncate_after=truncate_after,
                max_missed_peaks=max_missed_peaks,
                threshold_scale=threshold_scale,
                feature=feature)
            all_fits.extend(feature_fits)
        return all_fits

    def finalize_fit(self, feature_fit, charge_carrier=PROTON, subtract=True,
                     detection_threshold=0.1, max_missed_peaks=1):
        nodes = []
        start_time, end_time = find_bounds(feature_fit, detection_threshold)
        feat_iter = FeatureSetIterator(
            feature_fit.features, start_time, end_time)
        base_tid = feature_fit.theoretical
        charge = feature_fit.charge
        abs_charge = abs(charge)
        for eid in feat_iter:
            cleaned_eid, tid, n_missing = conform_envelopes(eid, base_tid.truncated_tid)
            rep_eid = drop_placeholders(cleaned_eid)
            n_real_peaks = len(rep_eid)
            if n_real_peaks == 0 or (n_real_peaks == 1 and abs_charge > 1) or \
               n_missing > max_missed_peaks:
                continue
            score = self.scorer.evaluate(None, cleaned_eid, tid)
            is_valid = True
            if np.isnan(score) or score < 0:
                is_valid = False
            envelope = [(e.mz, min(e.intensity, t.intensity)) for e, t in zip(cleaned_eid, tid)]
            if is_valid:
                total_abundance = sum(p[1] for p in envelope)
                monoisotopic_mass = neutral_mass(
                    base_tid.monoisotopic_mz, charge, charge_carrier=charge_carrier)
                reference_peak = first_peak(cleaned_eid)

                dpeak = DeconvolutedPeak(
                    neutral_mass=monoisotopic_mass, intensity=total_abundance,
                    charge=charge,
                    signal_to_noise=mean(p.signal_to_noise for p in rep_eid),
                    index=reference_peak.index,
                    full_width_at_half_max=mean(p.full_width_at_half_max for p in rep_eid),
                    a_to_a2_ratio=a_to_a2_ratio(tid),
                    most_abundant_mass=neutral_mass(
                        most_abundant_mz(cleaned_eid), charge, charge_carrier=charge_carrier),
                    average_mass=neutral_mass(
                        average_mz(cleaned_eid), charge, charge_carrier=charge_carrier),
                    score=score,
                    envelope=envelope,
                    mz=base_tid.monoisotopic_mz,
                    area=sum(e.area for e in cleaned_eid))

                time = feat_iter.current_time
                precursor_info_set = []
                for peak in rep_eid:
                    pinfo = self.precursor_map.precursor_for_peak((time, peak.index))
                    if pinfo is not None:
                        precursor_info_set.append(pinfo)

                node = DeconvolutedLCMSFeatureTreeNode(time, [dpeak], precursor_info_set)
                nodes.append(node)
            if subtract:
                for fpeak, tpeak in zip(cleaned_eid, envelope):
                    # If a theoretical peak uses up more than 70%
                    # of the abundance of a single peak, this peak
                    # should not contribute meaninfully to any other
                    # fits from now on. Set it's abundance to 1.0 as
                    # if it were fully used up.
                    ruin = (fpeak.intensity * 0.7) < tpeak[1]
                    if ruin:
                        fpeak.intensity = 1.0
                    else:
                        fpeak.intensity -= tpeak[1]
                    if fpeak.intensity < 0:
                        fpeak.intensity = 1.0
        for feature in feature_fit.features:
            if feature is None or isinstance(feature, EmptyFeature):
                continue
            feature.invalidate()
        if len(nodes) < self.minimum_size:
            return None

        result_feature = DeconvolutedLCMSFeature(
            nodes, feature_fit.charge,
            score=feature_fit.score, n_features=len(feature_fit),
            supporters=feature_fit.supporters)

        return result_feature

    def _clean_solutions(self, solutions):
        out = []
        for solution in solutions:
            parts = solution.split_sparse(self.maximum_time_gap)
            for part in parts:
                if len(part) < self.minimum_size:
                    continue
                out.append(part)
        return out

    def _merge_isobaric(self, solutions):
        if len(solutions) == 0:
            return []
        solutions.sort(key=lambda x: (x.neutral_mass, x.charge))
        last = solutions[0]
        out = []
        for solution in solutions[1:]:
            if last.neutral_mass == solution.neutral_mass and last.charge == solution.charge:
                last.sum(solution)
            else:
                out.append(last)
                last = solution
        out.append(last)
        return out

    def remove_peaks_below_threshold(self, minimum_intensity):
        out = []
        for feature in self.feature_map:
            keep = []
            for node in feature:
                if node.total_intensity() > minimum_intensity:
                    keep.append(node)
                else:
                    self.orphaned_nodes.append(node)

            if keep:
                filtered_feature = LCMSFeature(keep, feature_id=feature.feature_id)
                out.extend(filtered_feature.split_sparse())
        self.feature_map = LCMSFeatureMap(out)

    def store_solutions(self, fits, charge_carrier=PROTON, subtract=True, detection_threshold=0.1,
                        max_missed_peaks=1):
        solutions = []
        for fit in fits:
            extracted = extract_fitted_region(
                fit, detection_threshold=detection_threshold)
            solution = self.finalize_fit(
                extracted, charge_carrier=charge_carrier, subtract=subtract,
                detection_threshold=detection_threshold, max_missed_peaks=max_missed_peaks)
            if solution is None:
                continue
            solutions.append(solution)
        return solutions

    def build_dependence_network(self):
        self.dependence_network = FeatureDependenceGraph(self.feature_map)

    def _map_precursors(self, error_tolerance):
        printer("\tConstructing Precursor Seeds")
        rt_map = RTMap(self.feature_map)
        cache = dict()
        seeds = set()

        for key, pinfo in (self.precursor_map.mapping.items()):
            time, ix = key
            try:
                candidates = cache[time]
            except KeyError:
                candidates = rt_map.rt_tree.contains_point(time)
                cache[time] = candidates
            mz = pinfo["mz"]
            hits = [c.members[0] for c in candidates if c.contains_mz(mz, error_tolerance)]
            seeds.update(hits)
        return seeds

    def _make_iterator_state(self, error_tolerance=2e-5, charge_range=(1, 8), left_search=1, right_search=0,
                             charge_carrier=PROTON, truncate_after=0.95, maxiter=10, minimum_intensity=100,
                             convergence=0.01, max_missed_peaks=1, threshold_scale=0.3, relfitter=None,
                             detection_threshold=0.1):
        state = FeatureDeconvolutionIterationState(
            self, error_tolerance, charge_range, left_search, right_search, charge_carrier,
            truncate_after, maxiter, minimum_intensity, convergence, max_missed_peaks, threshold_scale,
            relfitter, detection_threshold)
        return state

    def deconvolute(self, error_tolerance=2e-5, charge_range=(1, 8), left_search=1, right_search=0,
                    charge_carrier=PROTON, truncate_after=0.95, maxiter=10, minimum_intensity=100,
                    convergence=0.01, max_missed_peaks=1, threshold_scale=0.3, relfitter=None,
                    detection_threshold=0.1):
        state = FeatureDeconvolutionIterationState(
            self, error_tolerance, charge_range, left_search, right_search, charge_carrier,
            truncate_after, maxiter, minimum_intensity, convergence, max_missed_peaks, threshold_scale,
            relfitter, detection_threshold)
        return state.run()


class FeatureDeconvolutionIterationState(object):
    def __init__(self, processor, error_tolerance=2e-5, charge_range=(1, 8), left_search=1, right_search=0,
                 charge_carrier=PROTON, truncate_after=0.95, maxiter=10, minimum_intensity=100,
                 convergence=0.01, max_missed_peaks=1, threshold_scale=0.3, relfitter=None,
                 detection_threshold=0.1):
        self.processor = processor
        self.last_signal_magnitude = sum(f.total_signal for f in self.processor.feature_map)
        self.next_signal_magnitude = 1.0
        self.total_signal_ratio = 1.0
        self.generation = None
        self.relations = None
        self.fits = None
        self.all_fits = None
        self.solutions = []
        self.disjoint_feature_clusters = None
        self.maxiter = maxiter
        self.iteration_count = 0

        self.error_tolerance = error_tolerance
        self.charge_range = charge_range
        self.left_search = left_search
        self.right_search = right_search
        self.charge_carrier = charge_carrier
        self.truncate_after = truncate_after
        self.minimum_intensity = minimum_intensity
        self.convergence = convergence
        self.max_missed_peaks = max_missed_peaks
        self.threshold_scale = threshold_scale
        self.relfitter = relfitter
        self.detection_threshold = detection_threshold
        self.debug = False

    def update_signal_ratio(self):
        self.next_signal_magnitude = sum(f.total_signal for f in self.processor.feature_map)
        self.total_signal_ratio = (self.last_signal_magnitude - self.next_signal_magnitude) / self.next_signal_magnitude
        printer("Signal Ratio: %0.3e (%0.3e, %0.3e)" % (
            self.total_signal_ratio, self.last_signal_magnitude, self.next_signal_magnitude))
        self.last_signal_magnitude = self.next_signal_magnitude
        return self.total_signal_ratio

    def setup(self):
        printer("Begin Iteration %d" % (self.iteration_count,))
        printer("Total Signal: %0.3e" % (sum(f.total_signal for f in self.processor.feature_map),))
        self.processor.remove_peaks_below_threshold(self.minimum_intensity)
        self.processor.build_dependence_network()

    def map_fits(self, report_interval=5):
        i = 0
        n = len(self.processor.feature_map)
        interval = int(max(n // report_interval, 1000))
        for feature in sorted(self.processor.feature_map, key=lambda x: x.mz, reverse=True):
            fits = self.processor.charge_state_determination(
                feature,
                error_tolerance=self.error_tolerance,
                charge_range=self.charge_range,
                left_search=self.left_search,
                right_search=self.right_search,
                charge_carrier=self.charge_carrier,
                truncate_after=self.truncate_after,
                max_missed_peaks=self.max_missed_peaks,
                threshold_scale=self.threshold_scale)
            if self.debug:
                print(feature, fits)
            i += 1
            if i % interval == 0:
                printer("\t%0.1f%%" % ((100. * i) / n,))
        self.all_fits = list(self.processor.dependence_network.dependencies)
        self.disjoint_feature_clusters = self.processor.dependence_network.find_non_overlapping_intervals()

        if self.relfitter is not None:
            self.relations = self.relfitter.fit(
                (d for cluster in self.disjoint_feature_clusters
                 for d in cluster), self.solutions)
            self.relfitter.predict((d for cluster in self.disjoint_feature_clusters
                                    for d in cluster))
        printer("\tExtracting Fits")
        self.fits = self.processor.select_best_disjoint_subgraphs(self.disjoint_feature_clusters)

    def postprocess(self, subtract=True, detection_threshold=0.1):
        self.generation = self.processor.store_solutions(
            self.fits, charge_carrier=self.charge_carrier,
            subtract=subtract, detection_threshold=detection_threshold,
            max_missed_peaks=self.max_missed_peaks)

        if self.relfitter is not None:
            self.relfitter.add_history((self.relations, self.generation))
        self.solutions.extend(self.generation)

    def step(self, report_interval=5, subtract=True):
        self.setup()

        self.map_fits(report_interval=report_interval)
        self.postprocess(
            subtract=subtract, detection_threshold=self.detection_threshold)

        self.iteration_count += 1
        self.update_signal_ratio()

        if self.total_signal_ratio < self.convergence:
            return True
        else:
            return False

    def run(self, callback=None):
        keep_going = True
        is_callable_callback = callable(callback)
        while keep_going:
            converged = self.step()
            if is_callable_callback:
                callback(self)
            if converged or self.iteration_count >= self.maxiter:
                keep_going = False
        self.solutions = self.processor._clean_solutions(self.solutions)
        return DeconvolutedLCMSFeatureMap(self.solutions)


def find_bounds(fit, detection_threshold=0.1, find_separation=True):
    start_time = 0
    end_time = float('inf')

    for f, p in zip(fit, fit.theoretical):
        if f is None:
            continue
        passed_threshold = p.intensity >= detection_threshold
        if f.start_time > start_time and passed_threshold:
            start_time = f.start_time
        if f.end_time < end_time and passed_threshold:
            end_time = f.end_time

    if fit.n_points > 0 and find_separation:
        last_score = float('inf')
        begin_i = 0
        end_i = len(fit.scores) - 1
        smoothed_scores = smooth_leveled(fit.times, fit.scores, 3)
        for i, score in enumerate(smoothed_scores):
            if score > 0 and last_score < 0:
                begin_i = i
            elif score < 0 and last_score > 0:
                end_i = i
            last_score = score
            pass
        if end_i < begin_i:
            end_i = len(fit.scores) - 1

        if start_time < fit.times[begin_i] and begin_i != 0:
            start_time = fit.times[begin_i]

        # if end_time > fit.times[end_i]:
        #     end_time = fit.times[end_i]

    return start_time, end_time


def extract_fitted_region(feature_fit, detection_threshold=0.1):
    fitted_features = []
    start_time, end_time = find_bounds(feature_fit, detection_threshold)
    for feature in feature_fit:
        if feature is None or isinstance(feature, EmptyFeature):
            fitted_features.append(feature)
            continue
        _, start_ix = feature.nodes.find_time(start_time)
        _, end_ix = feature.nodes.find_time(end_time)
        if feature.nodes[end_ix].time < end_time:
            end_ix += 1
        fitted = LCMSFeature(feature.nodes[start_ix:(end_ix + 1)], feature_id=feature.feature_id)
        if len(fitted) == 0:
            fitted_features.append(EmptyFeature(fitted.mz))
        else:
            fitted_features.append(fitted)
    start_ix = binsearch(feature_fit.times, start_time)
    end_ix = binsearch(feature_fit.times, end_time)
    extract = LCMSFeatureSetFit(
        fitted_features,
        feature_fit.theoretical,
        feature_fit.score,
        feature_fit.charge,
        feature_fit.missing_features,
        supporters=feature_fit.supporters,
        data=feature_fit.data,
        neutral_mass=feature_fit.neutral_mass,
        scores=feature_fit.scores[start_ix:end_ix + 1],
        times=feature_fit.times[start_ix:end_ix + 1])
    return extract


class RTFeatureNode(Interval):
    def __init__(self, feature):
        super(RTFeatureNode, self).__init__(
            feature.start_time, feature.end_time, [feature], mz=feature.mz)
        self.mz = feature.mz

    def contains_mz(self, mz, time, error_tolerance=2e-5):
        return abs((self.mz - mz) / mz) < error_tolerance


class RTMap(object):
    def __init__(self, features):
        self.rt_tree = IntervalTreeNode.build(map(RTFeatureNode, features))

    def locate_precursor_candidates(self, precursor_info):
        time = precursor_info.precursor.scan_time
        candidates = self.rt_tree.contains_point(time)
        return candidates

    def find_precursor(self, precursor_info, error_tolerance=2e-5):
        time = precursor_info.precursor.scan_time
        candidates = self.rt_tree.contains_point(time)
        matches = []
        for candidate in candidates:
            if candidate.contains_mz(precursor_info.mz, time, error_tolerance):
                matches.append(candidate)
        return matches

    def find_features_at_time(self, time):
        hits = self.rt_tree.contains_point(time)
        return [
            node.members[0] for node in hits
        ]


try:
    has_c = True
    _conform_envelopes = conform_envelopes
except ImportError:
    has_c = False



import unittest

from ms_deisotope.averagine import (
    peptide, calculate_mass, average_compositions,
    _Averagine, Averagine, add_compositions,
    AveragineCache, _AveragineCache, TheoreticalIsotopicPattern,
    _TheoreticalIsotopicPattern)


tid1 = [
    (1000.00, 0.587117149327, 1),
    (1001.0029404, 0.314398842767, 1),
    (1002.005551, 0.0984840079067, 1)
]

tid2 = [
    (1000.0, 0.312953555329, 2),
    (1000.5014313042954, 0.339021408804, 2),
    (1001.002474283496, 0.213306639087, 2),
    (1001.503389167319, 0.0983516834107, 2),
    (1002.00427365722, 0.0363667133695, 2)
]

composition = peptide.base_composition


def make_averagine_suite(averagine_class):
    _peptide = averagine_class(composition)

    class TestAveragine(unittest.TestCase):
        def test_isotopic_cluster(self):
            tid = _peptide.isotopic_cluster(1000, 1)
            for i, peak in enumerate(tid):
                self.assertAlmostEqual(peak.mz, tid1[i][0], 3)
                self.assertAlmostEqual(peak.intensity, tid1[i][1], 3)
                self.assertEqual(peak.charge, 1)

            tid = _peptide.isotopic_cluster(1000, 2)
            for peak, match in zip(tid, tid2):
                self.assertAlmostEqual(peak.mz, match[0], 3)
                self.assertAlmostEqual(peak.intensity, match[1], 3)
                self.assertEqual(peak.charge, match[2])

        def test_truncate_after(self):
            tid = _peptide.isotopic_cluster(1000, 1, truncate_after=0.95)
            inst = _peptide.isotopic_cluster(1000, 1, truncate_after=1.0).truncate_after(0.95)

            for i, p in enumerate(inst):
                self.assertAlmostEqual(tid[i].mz, p.mz, 3)
                self.assertAlmostEqual(tid[i].intensity, p.intensity)

        def __repr__(self):
            r = super(TestAveragine, self).__repr__()
            return "%s(%s)" % (r, averagine_class)

    return TestAveragine


TestPurePythonAveragine = make_averagine_suite(_Averagine)
TestAveragine = make_averagine_suite(Averagine)
TestAveragineCache = make_averagine_suite(AveragineCache)
TestPurePythonAveragineCache = make_averagine_suite(_AveragineCache)


class TestSupportMethods(unittest.TestCase):
    def test_average_composition(self):
        avgd = average_compositions([composition, composition])
        for k, v in avgd.items():
            self.assertAlmostEqual(v, composition[k], 3)

    def test_add_composition(self):
        avgd = add_compositions({}, composition)
        for k, v in avgd.items():
            self.assertAlmostEqual(v, composition[k], 3)


if __name__ == '__main__':
    unittest.main()



'''A collection of base classes for "exhaustive" search strategies.
These strategies attempt to assign *every* peak, but they are not meant
to be used on their own.

For complete implementations see :class:`~.AveragineDeconvoluter` and
:class:`~.AveraginePeakDependenceGraphDeconvoluter`.
'''

import operator

from ms_deisotope.constants import (
    ERROR_TOLERANCE, TRUNCATE_AFTER, IGNORE_BELOW, MAX_ITERATION,
    CONVERGENCE)

from ms_deisotope.peak_set import (
    DeconvolutedPeak,
    DeconvolutedPeakSet)

from ms_deisotope.averagine import PROTON, neutral_mass

from ms_deisotope.envelope_statistics import (
    average_mz,
    a_to_a2_ratio,
    most_abundant_mz)

from ms_deisotope.utils import (
    TrivialTargetedDeconvolutionResult)

from ms_deisotope.peak_dependency_network import (
    PeakDependenceGraph,
    NetworkedTargetedDeconvolutionResult)


from .base import DeconvoluterBase
from .utils import (
    ChargeIterator,
    has_previous_peak_at_charge,
    has_successor_peak_at_charge,
    drop_placeholders,
    count_placeholders,
    first_peak,
    mean,
    info,
    debug)


class ExhaustivePeakSearchDeconvoluterBase(DeconvoluterBase):
    """Provides common methods for algorithms which attempt to find a deconvolution for every peak
    in a spectrum. This assumes no dependence between different peaks, instead it relies on subtraction,
    breadth of search, and order of encounter to avoid artefactual fits. This is usually not reasonable,
    so instead please use this class's extension, :class:`PeakDependenceGraphDeconvoluterBase` which can
    express dependence of fits on common resources.

    This class is not meant to be instantiated, but instead used as a mixin for classes that also
    inherit from :class:`DeconvoluterBase` and provide methods `fit_theoretical_distribution`
    and `_fit_peaks_at_charges`

    Attributes
    ----------
    use_quick_charge: bool
        Whether or not to use the :title-reference:`QuickCharge` algorithm when generating
        putative charge states.
    incremental_truncation: float or :const:`None`
        If not :const:`None`, the isotopic pattern truncation lower bound to contract each fit to
        incrementally, generating new fits for each dropped peak using
        :meth:`~.DeconvoluterBase.fit_incremental_truncation`

    """

    def __init__(self, peaklist, *args, **kwargs):
        super(ExhaustivePeakSearchDeconvoluterBase,
              self).__init__(peaklist, *args, **kwargs)
        self.use_quick_charge = kwargs.get("use_quick_charge", False)
        self.incremental_truncation = kwargs.get("incremental_truncation", None)

    def _get_all_peak_charge_pairs(self, peak, error_tolerance=ERROR_TOLERANCE, charge_range=(1, 8),
                                   left_search_limit=3, right_search_limit=3,
                                   recalculate_starting_peak=True, use_quick_charge=False):
        """Construct the set of all unique candidate (monoisotopic peak, charge state) pairs using
        the provided search parameters.

        The search is performed using :func:`has_previous_peak_at_charge`, :func:`has_successor_peak_at_charge`,
        :meth:`_find_previous_putative_peak`, and :meth:`_find_next_putative_peak`.

        Parameters
        ----------
        peak : :class:`~.FittedPeak`
            The peak to start the search from
        error_tolerance : float, optional
            The parts-per-million error tolerance in m/z to search with. Defaults to |ERROR_TOLERANCE|
        charge_range : tuple, optional
            The range of charge states to consider. Defaults to (1, 8)
        left_search_limit : int, optional
            The number of steps to search to the left of `peak`. Defaults to 3
        right_search_limit : int, optional
            The number of steps to search to the right of `peak`. Defaults to 3
        recalculate_starting_peak : bool, optional
            Whether or not to re-calculate the putative starting peak m/z based upon nearby
            peaks close to where isotopic peaks for `peak` should be. Defaults to True

        Returns
        -------
        set
            The set of all unique candidate (monoisotopic peak, charge state)
        """

        target_peaks = set()

        charges = ChargeIterator(*charge_range)

        if use_quick_charge:
            charges.sequence_from_quickcharge(self.peaklist, peak)

        for charge in charges:
            target_peaks.add((peak, charge))

            # Look Left
            for i in range(1, left_search_limit):
                prev_peak = has_previous_peak_at_charge(
                    self, peak, charge, i)
                if prev_peak is None:
                    continue
                target_peaks.add((prev_peak, charge))

                if recalculate_starting_peak:
                    target_peaks.update(self._find_previous_putative_peak(
                        peak.mz, charge, i, 2 * error_tolerance))

            # Look Right
            for i in range(1, right_search_limit):
                nxt_peak = has_successor_peak_at_charge(
                    self, peak, charge, i)
                if nxt_peak is None:
                    continue
                target_peaks.add((nxt_peak, charge))

                if recalculate_starting_peak:
                    target_peaks.update(self._find_next_putative_peak(
                        peak.mz, charge, i, 2 * error_tolerance))

            if recalculate_starting_peak:
                for i in range(min(left_search_limit, 2)):
                    target_peaks.update(self._find_next_putative_peak(
                        peak.mz, charge, step=i, tolerance=2 * error_tolerance))

        return target_peaks

    def _fit_all_charge_states(self, peak, error_tolerance=ERROR_TOLERANCE, charge_range=(1, 8), left_search_limit=3,
                               right_search_limit=3, recalculate_starting_peak=True, charge_carrier=PROTON,
                               truncate_after=TRUNCATE_AFTER, ignore_below=IGNORE_BELOW):
        """Carry out the fitting process for `peak`.

        This method calls :meth:`_get_all_peak_charge_pairs` to collect all hypothetical solutions
        for `peak`, and invokes :meth:`_fit_peaks_at_charges` to evaluate them.

        The method :meth:`_fit_peaks_at_charges` is required by this interface, but is not defined by
        it, as it depends upon the underlying isotopic pattern fitting algorithm. See one of the
        Averagine-based algorithms for an implementation, such as :class:`AveragineDeconvoluterBase`,
        a complementary ancestor with this class.

        Parameters
        ----------
        peak : :class:`~.FittedPeak`
            The peak to start the search from
        error_tolerance : float, optional
            The parts-per-million error tolerance in m/z to search with. Defaults to |ERROR_TOLERANCE|
        charge_range : tuple, optional
            The range of charge states to consider. Defaults to (1, 8)
        left_search_limit : int, optional
            The number of steps to search to the left of `peak`. Defaults to 3
        right_search_limit : int, optional
            The number of steps to search to the right of `peak`. Defaults to 3
        recalculate_starting_peak : bool, optional
            Whether or not to re-calculate the putative starting peak m/z based upon nearby
            peaks close to where isotopic peaks for `peak` should be. Defaults to True
        charge_carrier : float, optional
            The mass of the charge carrier. Defaults to |PROTON|
        truncate_after : float, optional
            The percent of intensity to ensure is included in a theoretical isotopic pattern
            starting from the monoisotopic peak. This will cause theoretical isotopic patterns
            to be truncated, excluding trailing peaks which do not contribute substantially to
            the overall shape of the isotopic pattern.

        Returns
        -------
        set
            The set of :class:`~.IsotopicFitRecord` instances produced
        """
        target_peaks = self._get_all_peak_charge_pairs(
            peak, error_tolerance=error_tolerance,
            charge_range=charge_range,
            left_search_limit=left_search_limit,
            right_search_limit=right_search_limit,
            recalculate_starting_peak=recalculate_starting_peak,
            use_quick_charge=self.use_quick_charge)

        results = self._fit_peaks_at_charges(
            target_peaks, error_tolerance, charge_carrier=charge_carrier, truncate_after=truncate_after,
            ignore_below=ignore_below)
        if self.incremental_truncation is not None:
            solutions = set()
            for result in results:
                solutions.update(self.fit_incremental_truncation(result, self.incremental_truncation))
            results = solutions
        return (results)

    def charge_state_determination(self, peak, error_tolerance=ERROR_TOLERANCE, charge_range=(1, 8),
                                   left_search_limit=3, right_search_limit=3,
                                   charge_carrier=PROTON, truncate_after=TRUNCATE_AFTER,
                                   ignore_below=IGNORE_BELOW):
        """Determine the optimal isotopic fit for `peak`, extracting it's charge state and monoisotopic peak.

        This method invokes :meth:`_fit_all_charge_states`, and then uses :attr:`scorer`'s `select` method to
        choose the optimal solution.

        Parameters
        ----------
        peak : :class:`~.FittedPeak`
            The peak to start the search from
        error_tolerance : float, optional
            The parts-per-million error tolerance in m/z to search with. Defaults to |ERROR_TOLERANCE|
        charge_range : tuple, optional
            The range of charge states to consider. Defaults to (1, 8)
        left_search_limit : int, optional
            The number of steps to search to the left of `peak`. Defaults to 3
        right_search_limit : int, optional
            The number of steps to search to the right of `peak`. Defaults to 3
        charge_carrier : float, optional
            The mass of the charge carrier. Defaults to |PROTON|
        truncate_after : float, optional
            The percent of intensity to ensure is included in a theoretical isotopic pattern
            starting from the monoisotopic peak. This will cause theoretical isotopic patterns
            to be truncated, excluding trailing peaks which do not contribute substantially to
            the overall shape of the isotopic pattern.

        Returns
        -------
        :class:`~.IsotopicFitRecord`
            The best scoring isotopic fit
        """
        results = self._fit_all_charge_states(
            peak, error_tolerance=error_tolerance, charge_range=charge_range,
            left_search_limit=left_search_limit, right_search_limit=right_search_limit,
            charge_carrier=charge_carrier, truncate_after=truncate_after,
            ignore_below=ignore_below)

        if self.verbose:
            info("Fits for %r", peak)
            for rec in sorted(results)[-10:]:
                info(rec)
        try:
            result = self.scorer.select(results)
            return result
        except ValueError:
            return None

    def _make_deconvoluted_peak(self, fit, charge_carrier):
        score, charge, eid, tid = fit
        rep_eid = drop_placeholders(eid)
        total_abundance = sum(p.intensity for p in rep_eid)
        monoisotopic_mass = neutral_mass(
            tid.monoisotopic_mz, charge, charge_carrier)
        reference_peak = first_peak(eid)

        dpeak = DeconvolutedPeak(
            neutral_mass=monoisotopic_mass, intensity=total_abundance, charge=charge,
            signal_to_noise=mean(p.signal_to_noise for p in rep_eid),
            index=reference_peak.index,
            full_width_at_half_max=mean(
                p.full_width_at_half_max for p in rep_eid),
            a_to_a2_ratio=a_to_a2_ratio(tid),
            most_abundant_mass=neutral_mass(
                most_abundant_mz(eid), charge),
            average_mass=neutral_mass(average_mz(eid), charge),
            score=score,
            envelope=[(p.mz, p.intensity) for p in eid],
            mz=tid.monoisotopic_mz, fit=fit,
            area=sum(e.area for e in eid))
        return dpeak

    def deconvolute_peak(self, peak, error_tolerance=ERROR_TOLERANCE, charge_range=(1, 8),
                         left_search_limit=3, right_search_limit=3, charge_carrier=PROTON,
                         truncate_after=TRUNCATE_AFTER, ignore_below=IGNORE_BELOW):
        """Perform a deconvolution for `peak`, generating a new :class:`ms_deisotope.peak_set.DeconvolutedPeak` instance
        corresponding to the optimal solution.

        This new peak has an m/z matching the monoisotopic peak of the pattern containing `peak`, and its intensity
        is the sum of all the matched peaks in its isotopic pattern. Its charge, isotopic fit, and other qualities
        are derived from the :class:`ms_deisotope.scoring.IsotopicFitRecord` instance corresponding to its best
        solution.

        Parameters
        ----------
        peak : :class:`~.FittedPeak`
            The peak to start the search from
        error_tolerance : float, optional
            The parts-per-million error tolerance in m/z to search with. Defaults to |ERROR_TOLERANCE|
        charge_range : tuple, optional
            The range of charge states to consider. Defaults to (1, 8)
        left_search_limit : int, optional
            The number of steps to search to the left of `peak`. Defaults to 3
        right_search_limit : int, optional
            The number of steps to search to the right of `peak`. Defaults to 3
        charge_carrier : float, optional
            The mass of the charge carrier, or more specifically, the moiety which is added for
            each incremental change in charge state. Defaults to |PROTON|
        truncate_after : float, optional
            The percent of intensity to ensure is included in a theoretical isotopic pattern
            starting from the monoisotopic peak. This will cause theoretical isotopic patterns
            to be truncated, excluding trailing peaks which do not contribute substantially to
            the overall shape of the isotopic pattern.

        Returns
        -------
        :class:`~.DeconvolutedPeak`
        """
        fit = self.charge_state_determination(
            peak, error_tolerance=error_tolerance,
            charge_range=charge_range,
            left_search_limit=left_search_limit, right_search_limit=right_search_limit,
            charge_carrier=charge_carrier, truncate_after=truncate_after,
            ignore_below=ignore_below)
        if fit is None:
            return
        tid = fit.theoretical
        dpeak = self._make_deconvoluted_peak(fit, charge_carrier)
        self._deconvoluted_peaks.append(dpeak)
        if self.use_subtraction:
            self.subtraction(tid, error_tolerance)
        return dpeak

    def targeted_deconvolution(self, peak, error_tolerance=ERROR_TOLERANCE, charge_range=(1, 8),
                               left_search_limit=3, right_search_limit=3,
                               charge_carrier=PROTON, truncate_after=TRUNCATE_AFTER,
                               ignore_below=IGNORE_BELOW):
        """Express the intent that this peak's deconvolution solution will be retrieved at a later point in the process
        and that it should be deconvoluted, and return a handle to retrieve the results with.

        This algorithm's implementation is simple enough that this is equivalent to just performing the deconvolution
        now and storing the result in a :class:`~.TrivialTargetedDeconvolutionResult` instance.

        Otherwise identical to :meth:`deconvolute_peak`.

        Parameters
        ----------
        peak : :class:`~.FittedPeak`
            The peak to start the search from
        error_tolerance : float, optional
            The parts-per-million error tolerance in m/z to search with. Defaults to |ERROR_TOLERANCE|
        charge_range : tuple, optional
            The range of charge states to consider. Defaults to (1, 8)
        left_search_limit : int, optional
            The number of steps to search to the left of `peak`. Defaults to 3
        right_search_limit : int, optional
            The number of steps to search to the right of `peak`. Defaults to 3
        charge_carrier : float, optional
            The mass of the charge carrier. Defaults to |PROTON|
        truncate_after : float, optional
            The percent of intensity to ensure is included in a theoretical isotopic pattern
            starting from the monoisotopic peak. This will cause theoretical isotopic patterns
            to be truncated, excluding trailing peaks which do not contribute substantially to
            the overall shape of the isotopic pattern.

        Returns
        -------
        :class:`~.TrivialTargetedDeconvolutionResult`
        """
        dpeak = self.deconvolute_peak(
            peak, error_tolerance=error_tolerance, charge_range=charge_range,
            left_search_limit=left_search_limit, right_search_limit=right_search_limit,
            charge_carrier=charge_carrier, truncate_after=truncate_after,
            ignore_below=ignore_below)
        result = TrivialTargetedDeconvolutionResult(self, dpeak, peak)
        return result

    def deconvolute(self, error_tolerance=ERROR_TOLERANCE, charge_range=(1, 8),
                    order_chooser=operator.attrgetter('index'),
                    left_search_limit=3, right_search_limit=3, charge_carrier=PROTON,
                    truncate_after=TRUNCATE_AFTER, ignore_below=IGNORE_BELOW,
                    **kwargs):
        """Completely deconvolute the spectrum.

        Visit each peak in the order chosen by `order_chooser`, and call :meth:`deconvolute_peak`
        on it with the provided arguments. This assumes all overlaps in isotopic pattern are captured
        by the search limits. This is usually not the case. For an alternative see
        :class:`PeakDependenceGraphDeconvoluterBase`

        Parameters
        ----------
        error_tolerance : float, optional
            The parts-per-million error tolerance in m/z to search with. Defaults to |ERROR_TOLERANCE|
        charge_range : tuple, optional
            The range of charge states to consider. Defaults to (1, 8)
        order_chooser : callable, optional:
            A callable used as a key function for sorting peaks into the order they will
            be visited during deconvolution. Defaults to :obj:`operator.attrgetter("index")`
        left_search_limit : int, optional
            The number of steps to search to the left of `peak`. Defaults to 3
        right_search_limit : int, optional
            The number of steps to search to the right of `peak`. Defaults to 3
        charge_carrier : float, optional
            The mass of the charge carrier. Defaults to |PROTON|
        truncate_after : float, optional
            The percent of intensity to ensure is included in a theoretical isotopic pattern
            starting from the monoisotopic peak. This will cause theoretical isotopic patterns
            to be truncated, excluding trailing peaks which do not contribute substantially to
            the overall shape of the isotopic pattern.

        Returns
        -------
        :class:`~.DeconvolutedPeakSet`
        """
        i = 0
        for peak in sorted(self.peaklist, key=order_chooser, reverse=True):
            if peak.mz < 2 or peak.intensity < self.minimum_intensity:
                continue
            self.deconvolute_peak(
                peak, error_tolerance=error_tolerance, charge_range=charge_range,
                left_search_limit=left_search_limit,
                right_search_limit=right_search_limit, charge_carrier=charge_carrier,
                truncate_after=truncate_after, ignore_below=ignore_below)
            i += 1

        if self.merge_isobaric_peaks:
            self._deconvoluted_peaks = self._merge_peaks(
                self._deconvoluted_peaks)

        return DeconvolutedPeakSet(self._deconvoluted_peaks).reindex()


try:
    from ms_deisotope._c.deconvoluter_base import (
        _get_all_peak_charge_pairs as _c_get_all_peak_charge_pairs,
        _make_deconvoluted_peak as _c_make_deconvoluted_peak)
    ExhaustivePeakSearchDeconvoluterBase._get_all_peak_charge_pairs = _c_get_all_peak_charge_pairs
    ExhaustivePeakSearchDeconvoluterBase._make_deconvoluted_peak = _c_make_deconvoluted_peak
except ImportError as e:
    print(e)


class PeakDependenceGraphDeconvoluterBase(ExhaustivePeakSearchDeconvoluterBase):
    """Extends the concept of :class:`ExhaustivePeakSearchDeconvoluterBase` to include a way to handle
    conflicting solutions which claim the same experimental peak.

    This lets the Deconvoluter assign a single peak only once, and to the "best" solution to use it. To
    do this, the Deconvoluter constructs a graph where peaks are nodes, and isotopic fits are hyperedges
    connecting multiple nodes. Rather than deconvoluting the spectrum step by step, assigning signal as
    it explores the spectrum, the Deconvoluter instead inserts each considered isotopic fit into the graph.
    After completely traversing the spectrum, the Deconvoluter solves the dependence graph attempting to
    maximize some criterion and produce a set of disjoint isotopic fits. These fits are then assigned signal
    and added to the deconvoluted spectrum as normal.

    The criterion used is currently a greedy maximization (or minimization) of each connected component of
    the peak dependence graph.

    Attributes
    ----------
    max_missed_peaks : int
        The maximum number of missing peaks to tolerate in an isotopic fit
    peak_dependency_network : :class:`~.PeakDependenceGraph`
        The peak dependence graph onto which isotopic fit dependences on peaks
        are constructed and solved.
    """

    def __init__(self, peaklist, *args, **kwargs):
        max_missed_peaks = kwargs.get("max_missed_peaks", 1)
        self.subgraph_solver_type = kwargs.get("subgraph_solver", 'top')
        super(PeakDependenceGraphDeconvoluterBase,
              self).__init__(peaklist, *args, **kwargs)
        self.peak_dependency_network = PeakDependenceGraph(
            self.peaklist, maximize=self.scorer.is_maximizing())
        self.max_missed_peaks = max_missed_peaks
        self.fit_postprocessor = kwargs.get("fit_postprocessor", None)
        self._priority_map = {}

    @property
    def max_missed_peaks(self):
        """The maximum number of missed peaks per isotopic fit record permitted.

        This property directly mirrors :attr:`PeakDependenceGraph.max_missed_peaks`

        Returns
        -------
        int
        """
        return self.peak_dependency_network.max_missed_peaks

    @max_missed_peaks.setter
    def max_missed_peaks(self, value):
        self.peak_dependency_network.max_missed_peaks = value

    def _explore_local(self, peak, error_tolerance=ERROR_TOLERANCE, charge_range=(1, 8), left_search_limit=1,
                       right_search_limit=0, charge_carrier=PROTON, truncate_after=TRUNCATE_AFTER,
                       ignore_below=IGNORE_BELOW):
        """Given a peak, explore the local neighborhood for candidate isotopic fits and add each
        fit above a threshold to the peak dependence graph.

        The threshold assumes that a single peak's neighborhood will contain many, many fits, but
        that only the top `n` scoring fits are worth considering. For now, `n` is fixed at 100 or
        the half number of fits returned, whichever is larger. This is to prevent the fit graph
        from growing out of control and wasting time storing impractical fits. Any fit added to
        the graph will have to pass :attr:`scorer.select` as well, so weak fits will never be added,
        regardless of how many fits are allowed to be inserted.

        Parameters
        ----------
        peak : :class:`~.FittedPeak`
            The peak to start the search from
        error_tolerance : float, optional
            The parts-per-million error tolerance in m/z to search with. Defaults to |ERROR_TOLERANCE|
        charge_range : tuple, optional
            The range of charge states to consider. Defaults to (1, 8)
        left_search_limit : int, optional
            The number of steps to search to the left of `peak`. Defaults to 1
        right_search_limit : int, optional
            The number of steps to search to the right of `peak`. Defaults to 0
        charge_carrier : float, optional
            The mass of the charge carrier. Defaults to |PROTON|
        truncate_after : float, optional
            The percent of intensity to ensure is included in a theoretical isotopic pattern
            starting from the monoisotopic peak. This will cause theoretical isotopic patterns
            to be truncated, excluding trailing peaks which do not contribute substantially to
            the overall shape of the isotopic pattern.

        Returns
        -------
        int
            The number of fits added to the graph
        """
        results = self._fit_all_charge_states(
            peak, error_tolerance=error_tolerance, charge_range=charge_range, left_search_limit=left_search_limit,
            charge_carrier=charge_carrier, truncate_after=truncate_after, ignore_below=ignore_below)

        hold = set()
        for fit in results:
            if fit.charge > 1 and len(drop_placeholders(fit.experimental)) == 1:
                continue
            hold.add(fit)

        results = hold
        n = len(results)

        stop = max(min(n // 2, 100), 10)
        if n == 0:
            return 0
        i = 0
        for i in range(stop):
            if len(results) == 0:
                break
            candidate = self.scorer.select(results)
            if candidate is None:
                break
            self.peak_dependency_network.add_fit_dependence(candidate)
            results.discard(candidate)

        return i

    def populate_graph(self, error_tolerance=ERROR_TOLERANCE, charge_range=(1, 8), left_search_limit=1,
                       right_search_limit=0, charge_carrier=PROTON,
                       truncate_after=TRUNCATE_AFTER, ignore_below=IGNORE_BELOW):
        """Visit each experimental peak and execute :meth:`_explore_local` on it with the provided
        parameters, populating the peak dependence graph with all viable candidates.

        Parameters
        ----------
        peak : :class:`~.FittedPeak`
        error_tolerance : float, optional
            The parts-per-million error tolerance in m/z to search with. Defaults to |ERROR_TOLERANCE|
        charge_range : tuple, optional
            The range of charge states to consider. Defaults to (1, 8)
        left_search_limit : int, optional
            The number of steps to search to the left of `peak`. Defaults to 1
        right_search_limit : int, optional
            The number of steps to search to the right of `peak`. Defaults to 0
        charge_carrier : float, optional
            The mass of the charge carrier. Defaults to |PROTON|
        truncate_after : float, optional
            The percent of intensity to ensure is included in a theoretical isotopic pattern
            starting from the monoisotopic peak. This will cause theoretical isotopic patterns
            to be truncated, excluding trailing peaks which do not contribute substantially to
            the overall shape of the isotopic pattern.
        """
        for peak in self.peaklist:
            if peak in self._priority_map or peak.intensity < self.minimum_intensity:
                if self.verbose:
                    debug("Skipping %r (%r)", peak,
                          peak.intensity < self.minimum_intensity)
                continue
            n = self._explore_local(
                peak, error_tolerance=error_tolerance, charge_range=charge_range,
                left_search_limit=left_search_limit, right_search_limit=right_search_limit,
                charge_carrier=charge_carrier,
                truncate_after=truncate_after, ignore_below=ignore_below)
            if self.verbose:
                debug("Exlporing Area Around %r Yielded %d Fits", peak, n)

    def postprocess_fits(self, error_tolerance=ERROR_TOLERANCE, charge_range=(1, 8),
                         charge_carrier=PROTON, *args, **kwargs):
        """Postprocesses fits before solving the peak dependence graph.

        Currently a no-op.

        Parameters
        ----------
        error_tolerance : float, optional
            The parts-per-million error tolerance in m/z to search with. Defaults to |ERROR_TOLERANCE|
        charge_range : tuple, optional
            The range of charge states to consider. Defaults to (1, 8)
        charge_carrier : float, optional
            The mass of the charge carrier. Defaults to |PROTON|
        """
        if self.fit_postprocessor is None:
            return

    def select_best_disjoint_subgraphs(self, error_tolerance=ERROR_TOLERANCE, charge_carrier=PROTON):
        """Construct connected envelope graphs from :attr:`peak_dependency_network` and
        extract the best disjoint isotopic pattern fits in each envelope graph. This in turn
        produces one or more :class:`DeconvolutedPeak` instances from each disjoint fit,
        which are processed and added to the results set.

        Parameters
        ----------
        error_tolerance : float, optional
            The error tolerance to use when performing subtraction, if subtraction is
            being performed.
        charge_carrier : float, optional
            The mass of the charge carrier as used for the deconvolution. Required to
            back-out the neutral mass of the deconvoluted result
        """
        disjoint_envelopes = self.peak_dependency_network.find_non_overlapping_intervals()
        i = 0
        if self.subgraph_solver_type == 'disjoint':
            solver = self._solve_subgraph_disjoint
        elif self.subgraph_solver_type == 'iterative':
            solver = self._solve_subgraph_iterative
        elif self.subgraph_solver_type == 'top':
            solver = self._solve_subgraph_top
        else:
            raise ValueError("Unknown solver type %r" %
                             (self.subgraph_solver_type, ))

        for cluster in disjoint_envelopes:
            solutions = solver(cluster, error_tolerance, charge_carrier)
            for dpeak in solutions:
                self.peak_dependency_network.add_solution(dpeak.fit, dpeak)
                self._deconvoluted_peaks.append(dpeak)
                i += 1

    def _solve_subgraph_top(self, cluster, error_tolerance=ERROR_TOLERANCE, charge_carrier=PROTON):
        """Given a :class:`~.DependenceCluster`, return the single best fit from the collection of
        co-dependent fits.

        Parameters
        ----------
        cluster : :class:`~.DependenceCluster`
            The connected subgraph whose nodes will be searched
        error_tolerance : float, optional
            The error tolerance to use when performing subtraction, if subtraction is
            being performed.
        charge_carrier : float, optional
            The mass of the charge carrier as used for the deconvolution. Required to
            back-out the neutral mass of the deconvoluted result

        Returns
        -------
        list of :class:`~DeconvolutedPeak`
            The solved deconvolution solutions
        """
        fit = cluster[0]
        _, _, eid, tid = fit
        rep_eid = drop_placeholders(eid)
        if len(rep_eid) == 0:
            return []
        dpeak = self._make_deconvoluted_peak(fit, charge_carrier)
        if self.use_subtraction:
            self.subtraction(tid, error_tolerance)
        return [dpeak]

    def _solve_subgraph_disjoint(self, cluster, error_tolerance=ERROR_TOLERANCE, charge_carrier=PROTON):
        """Given a :class:`~.DependenceCluster`, find a greedy disjoint set of isotopic fits.

        Parameters
        ----------
        cluster : :class:`~.DependenceCluster`
            The connected subgraph whose nodes will be searched
        error_tolerance : float, optional
            The error tolerance to use when performing subtraction, if subtraction is
            being performed.
        charge_carrier : float, optional
            The mass of the charge carrier as used for the deconvolution. Required to
            back-out the neutral mass of the deconvoluted result

        Returns
        -------
        list of :class:`~DeconvolutedPeak`
            The solved deconvolution solutions

        """
        disjoint_best_fits = cluster.disjoint_best_fits()
        i = 0
        solutions = []
        for fit in disjoint_best_fits:
            _, _, eid, tid = fit
            rep_eid = drop_placeholders(eid)
            if len(rep_eid) == 0:
                continue
            dpeak = self._make_deconvoluted_peak(fit, charge_carrier)
            solutions.append(dpeak)
            i += 1
            if self.use_subtraction:
                self.subtraction(tid, error_tolerance)
        return solutions

    def _solve_subgraph_iterative(self, cluster, error_tolerance=ERROR_TOLERANCE, charge_carrier=PROTON):
        """Given a :class:`~.DependenceCluster`, build a :class:`~.ConnectedSubgraph` and incrementally
        subtract the best fitting solution and update its overlapping envelopes.

        Parameters
        ----------
        cluster : :class:`~.DependencyCluster`
            The connected subgraph whose nodes will be searched
        error_tolerance : float, optional
            The error tolerance to use when performing subtraction, if subtraction is
            being performed.
        charge_carrier : float, optional
            The mass of the charge carrier as used for the deconvolution. Required to
            back-out the neutral mass of the deconvoluted result

        Returns
        -------
        list of :class:`~DeconvolutedPeak`
            The solved deconvolution solutions
        """
        print(cluster)
        subgraph = cluster.build_graph()
        solutions = []
        mask = set()
        best_node = subgraph[0]
        peak = self._make_deconvoluted_peak(best_node.fit, charge_carrier)
        solutions.append(peak)
        if self.use_subtraction:
            self.subtraction(best_node.fit.theoretical, error_tolerance)
        mask.add(best_node.index)
        print("Masking %d (%0.3f, %d)" % (
            best_node.index, best_node.fit.monoisotopic_peak.mz, best_node.fit.charge))

        overlapped_nodes = list(best_node.overlap_edges)
        maximize = subgraph.maximize

        n = len(subgraph)
        while len(mask) != n:
            best_node = None
            best_score = 0 if maximize else float('inf')
            retained = []
            for node in overlapped_nodes:
                if node.index in mask:
                    continue
                missed_peaks = node.fit.missed_peaks = count_placeholders(
                    node.fit.experimental)
                total_peaks = len(node.fit.experimental)
                invalid_peak_count = (missed_peaks >= total_peaks - 1 and abs(node.fit.charge) > 1
                                     ) or missed_peaks == total_peaks
                if invalid_peak_count or missed_peaks > self.max_missed_peaks:
                    mask.add(node.index)
                    print("Masking %d (%0.3f, %d). Invalidated Peak Count %r | Missed Peaks %d" % (
                        node.index, node.fit.monoisotopic_peak.mz,
                        node.fit.charge, invalid_peak_count, missed_peaks))

                    continue
                fit = node.fit
                fit.theoretical.normalize().scale(fit.experimental, self.scale_method)
                fit.score = self.scorer.evaluate(
                    self.peaklist, fit.experimental, fit.theoretical.peaklist)
                if self.scorer.reject(fit):
                    mask.add(node.index)
                    continue
                else:
                    retained.append(node)
                if maximize:
                    if fit.score > best_score:
                        best_node = node
                        best_score = fit.score
                else:
                    if fit.score < best_score:
                        best_node = node
                        best_score = fit.score

            if best_node is not None:
                peak = self._make_deconvoluted_peak(
                    best_node.fit, charge_carrier)
                solutions.append(peak)
                if self.use_subtraction:
                    self.subtraction(
                        best_node.fit.theoretical, error_tolerance)
                mask.add(best_node.index)
                print("Masking %d (%0.3f, %d)" % (
                    best_node.index, best_node.fit.monoisotopic_peak.mz, best_node.fit.charge))

                overlapped_nodes = [
                    node for node in best_node.overlap_edges if node.index not in mask]
            else:
                overlapped_nodes = []
            if not overlapped_nodes and len(mask) != n:
                overlapped_nodes = [
                    node for node in subgraph if node.index not in mask]
        return solutions

    def targeted_deconvolution(self, peak, error_tolerance=ERROR_TOLERANCE, charge_range=(1, 8),
                               left_search_limit=3, right_search_limit=3, charge_carrier=PROTON,
                               truncate_after=TRUNCATE_AFTER, ignore_below=IGNORE_BELOW):
        """Express the intent that this peak's deconvolution solution will be retrieved at a later point in the process
        and that it should be deconvoluted, and return a handle to retrieve the results with.

        As the operation does not immediately result in a deconvoluted peak but just adds the resulting fits to
        :attr:`peak_dependency_network`, this method constructs an instance of
        :class:`~.NetworkedTargetedDeconvolutionResult` which holds all
        the required information for recovering the best fit containing `peak`.

        Parameters
        ----------
        peak : :class:`~.FittedPeak`
            The peak to start the search from
        error_tolerance : float, optional
            The parts-per-million error tolerance in m/z to search with. Defaults to |ERROR_TOLERANCE|
        charge_range : tuple, optional
            The range of charge states to consider. Defaults to (1, 8)
        left_search_limit : int, optional
            The number of steps to search to the left of `peak`. Defaults to 3
        right_search_limit : int, optional
            The number of steps to search to the right of `peak`. Defaults to 3
        charge_carrier : float, optional
            The mass of the charge carrier. Defaults to |PROTON|

        Returns
        -------
        :class:`~.NetworkedTargetedDeconvolutionResult`
        """
        self._explore_local(
            peak, error_tolerance=error_tolerance, charge_range=charge_range,
            left_search_limit=left_search_limit,
            right_search_limit=right_search_limit, charge_carrier=charge_carrier,
            truncate_after=truncate_after, ignore_below=ignore_below)
        result = NetworkedTargetedDeconvolutionResult(self, peak)
        self._priority_map[peak] = result
        return result

    def deconvolute(self, error_tolerance=ERROR_TOLERANCE, charge_range=(1, 8),   # pylint: disable=arguments-differ
                    left_search_limit=1, right_search_limit=0, iterations=MAX_ITERATION,
                    charge_carrier=PROTON, truncate_after=TRUNCATE_AFTER, ignore_below=IGNORE_BELOW,
                    convergence=CONVERGENCE, **kwargs):
        """Completely deconvolute the spectrum.

        For each iteration, clear :attr:`peak_depencency_network`, then invoke :meth:`populate_graph`
        followed by :meth:`select_best_disjoint_subgraphs` to populate the resulting
        :class:`~.DeconvolutedPeakSet`

        Parameters
        ----------
        error_tolerance : float, optional
            The parts-per-million error tolerance in m/z to search with. Defaults to |ERROR_TOLERANCE|
        charge_range : tuple, optional
            The range of charge states to consider. Defaults to (1, 8)
        order_chooser : callable, optional:
            A callable used as a key function for sorting peaks into the order they will
            be visited during deconvolution. Defaults to :obj:`operator.attrgetter("index")`
        left_search_limit : int, optional
            The number of steps to search to the left of :obj:`peak`. Defaults to 1
        right_search_limit : int, optional
            The number of steps to search to the right of :obj:`peak`. Defaults to 0
        charge_carrier : float, optional
            The mass of the charge carrier. Defaults to |PROTON|
        truncate_after : float, optional
            The percent of intensity to ensure is included in a theoretical isotopic pattern
            starting from the monoisotopic peak. This will cause theoretical isotopic patterns
            to be truncated, excluding trailing peaks which do not contribute substantially to
            the overall shape of the isotopic pattern.
        ignore_below : float, optional
            The minimum relative abundance to consider a peak in a theoretical isotopic
            pattern
        convergence : float, optional
            The threshold of the below which after the `(sum(intensity_before) - sum(
            intensity_after)) / sum(intensity_after)`

        Returns
        -------
        :class:`~.DeconvolutedPeakSet`
        """
        if not self.use_subtraction:
            iterations = 1

        begin_signal = sum([p.intensity for p in self.peaklist])
        i = 0
        for i in range(iterations):
            if self.verbose:
                info("<== Starting Iteration %d ===================>", (i, ))
            # The first iteration doesn't need to have the entire graph
            # rebuilt and might have been seeded with targeted queries
            if i != 0:
                self.peak_dependency_network.reset()
            self.populate_graph(
                error_tolerance=error_tolerance, charge_range=charge_range,
                left_search_limit=left_search_limit, right_search_limit=right_search_limit,
                charge_carrier=charge_carrier,
                truncate_after=truncate_after, ignore_below=ignore_below)
            self.postprocess_fits(
                charge_range=charge_range,
                charge_carrier=charge_carrier,
                error_tolerance=error_tolerance)
            self.select_best_disjoint_subgraphs(
                error_tolerance, charge_carrier)
            self._slice_cache.clear()
            end_signal = sum([p.intensity for p in self.peaklist]) + 1

            if (begin_signal - end_signal) / end_signal < convergence:
                if self.verbose:
                    info("(%0.4e - %0.4e) / %0.4e < %0.2g, Converged!",
                         begin_signal, end_signal, end_signal, convergence)
                break
            begin_signal = end_signal
        else:
            if self.verbose:
                info("Did Not Converge.")
        if self.verbose:
            info("Finished Deconvolution in %d Iterations", (i + 1, ))
        if self.merge_isobaric_peaks:
            self._deconvoluted_peaks = self._merge_peaks(
                self._deconvoluted_peaks)

        return DeconvolutedPeakSet(list(self._deconvoluted_peaks)).reindex()


try:
    from ms_deisotope._c.deconvoluter_base import (
        _explore_local as _c_explore_local,
        populate_graph as cpopulate_graph)
    PeakDependenceGraphDeconvoluterBase._explore_local = _c_explore_local
    PeakDependenceGraphDeconvoluterBase.populate_graph = cpopulate_graph
except ImportError as e:
    print(e)



import base64
import zlib

import numpy as np

from ms_deisotope.peak_set import Envelope, EnvelopePair

COMPRESSION_NONE = 'none'
COMPRESSION_ZLIB = 'zlib'


def encode_array(array, compression=COMPRESSION_NONE, dtype=np.float32):
    bytestring = np.asanyarray(array).astype(dtype).tobytes()
    if compression == COMPRESSION_NONE:
        bytestring = bytestring
    elif compression == COMPRESSION_ZLIB:
        bytestring = zlib.compress(bytestring)
    else:
        raise ValueError("Unknown compression: %s" % compression)
    encoded_string = base64.standard_b64encode(bytestring)
    return encoded_string


def decode_array(bytestring, compression=COMPRESSION_NONE, dtype=np.float32):
    try:
        decoded_string = bytestring.encode("ascii")
    except AttributeError:
        decoded_string = bytestring
    decoded_string = base64.decodestring(decoded_string)
    if compression == COMPRESSION_NONE:
        decoded_string = decoded_string
    elif compression == COMPRESSION_ZLIB:
        decoded_string = zlib.decompress(decoded_string)
    else:
        raise ValueError("Unknown compression: %s" % compression)
    array = np.fromstring(decoded_string, dtype=dtype)
    return array


def envelopes_to_array(envelope_list, dtype=np.float32):
    collection = []
    for envelope in envelope_list:
        collection.append(0)
        collection.append(0)
        for pair in envelope:
            collection.extend(pair)
    return np.array(collection, dtype=np.float32)


def decode_envelopes(array):
    envelope_list = []
    current_envelope = []
    i = 0
    n = len(array)
    while i < n:
        a = array[i]
        b = array[i + 1]
        i += 2
        if a == 0 and b == 0:
            if current_envelope is not None:
                if current_envelope:
                    envelope_list.append(Envelope(current_envelope))
                current_envelope = []
        else:
            current_envelope.append(EnvelopePair(a, b))
    envelope_list.append(Envelope(current_envelope))
    return envelope_list


try:
    has_c = True
    _decode_envelopes = decode_envelopes
    from ms_deisotope._c.utils import decode_envelopes
except ImportError:
    has_c = False



cdef class SpanningMixin(object):
    cdef:
        public double start
        public double end

    cdef bint _contains(self, double i)

    cpdef bint contains(self, double i)

    cpdef bint overlaps(self, SpanningMixin interval)

    cpdef double overlap_size(self, SpanningMixin interval)

    cpdef bint is_contained_in_interval(self, SpanningMixin interval)

    cpdef bint contains_interval(self, SpanningMixin interval)


cdef class Interval(SpanningMixin):
    cdef:
        public object members
        public dict data


cdef class IntervalTreeNode(object):
    cdef:
        public double center
        public double start
        public double end
        public IntervalTreeNode left
        public IntervalTreeNode right
        public list contained
        public int level
        public IntervalTreeNode parent

    cpdef list contains_point(self, double x)
    cpdef list _overlaps_interval(self, double start, double end)
    cpdef list overlaps(self, double start, double end)
    cpdef bint _eq(self, IntervalTreeNode other)



from ms_peak_picker._c.peak_set cimport FittedPeak, PeakBase


cdef class LCMSFeatureTreeList(object):
    cdef:
        public list roots
        public set _node_id_hash

    @staticmethod
    cdef LCMSFeatureTreeList _create(list roots)

    cdef void _invalidate(self)

    cpdef LCMSFeatureTreeNodeBase _make_node(self, double time, list peaks)

    cpdef tuple find_time(self, double time)
    cdef inline LCMSFeatureTreeNodeBase _find_time(self, double time, size_t* indexout)
    cdef inline LCMSFeatureTreeNodeBase getitem(self, size_t i)

    cdef inline size_t get_size(self)


cdef class LCMSFeatureTreeNodeBase(object):
    cdef:
        public double time
        public list members
        public PeakBase _most_abundant_member
        public object node_id

    cpdef double _total_intensity_members(self)
    cpdef double max_intensity(self)
    cpdef double total_intensity(self)

    cdef PeakBase _find_most_abundant_member(self)
    cpdef _calculate_most_abundant_member(self)
    cpdef _recalculate(self)

    cdef PeakBase getitem(self, size_t i)
    cdef inline size_t get_members_size(self)




cdef class LCMSFeatureTreeNode(LCMSFeatureTreeNodeBase):
    cdef:

        public double _mz

    cpdef bint _eq(self, LCMSFeatureTreeNode other)
    cpdef bint _ne(self, LCMSFeatureTreeNode other)
    cdef inline FittedPeak getpeak(self, size_t i)

    cdef double get_mz(self)


cdef class FeatureBase(object):
    cdef:
        public LCMSFeatureTreeList nodes
        public double _mz
        public double _start_time
        public double _end_time

    cdef double get_mz(self)
    cdef double get_start_time(self)
    cdef double get_end_time(self)

    cdef inline size_t get_size(self)
    cpdef tuple find_time(self, double time)
    cdef inline LCMSFeatureTreeNode _find_time(self, double time, size_t* indexout)


cdef class LCMSFeature(FeatureBase):
    cdef:
        public double _total_intensity
        public double _last_mz
        public object _times
        public object _peaks
        public object adducts
        public object used_as_adduct
        public object created_at
        public object feature_id
        public RunningWeightedAverage _peak_averager

    cpdef bint _eq(self, other)
    cpdef bint _ne(self, other)
    cpdef bint overlaps_in_time(self, FeatureBase interval)

    cdef LCMSFeatureTreeNode getitem(self, size_t i)
    cdef void _feed_peak_averager(self)

    cpdef insert_node(self, LCMSFeatureTreeNode node)
    cpdef insert(self, PeakBase peak, double time)
    cpdef _invalidate(self, bint reaverage=*)


cdef class EmptyFeature(FeatureBase):
    @staticmethod
    cdef EmptyFeature _create(double mz)


cdef class FeatureSetIterator(object):
    cdef:
        public list features
        public list real_features
        public double start_time
        public double end_time
        public double last_time_seen
        size_t* index_list

        @staticmethod
        cdef FeatureSetIterator _create(list features)

        @staticmethod
        cdef FeatureSetIterator _create_with_threshold(list features, list theoretical_distribution, double detection_threshold)

        cdef void _initialize(self, list features)

        cpdef init_indices(self)
        cpdef double get_next_time(self)
        cpdef double get_current_time(self)
        cpdef bint has_more(self)
        cpdef list get_peaks_for_time(self, double time)
        cpdef list get_next_value(self)

        cdef inline size_t get_size(self)
        cdef inline FeatureBase getitem(self, size_t i)


cdef class RunningWeightedAverage(object):
    cdef:
        public list accumulator
        public double current_mean
        public size_t current_count
        public double total_weight

    cpdef add(self, PeakBase peak)
    cpdef double recompute(self)
    cpdef RunningWeightedAverage update(self, iterable)
    cpdef double _bootstrap(self, size_t n=*, size_t k=*)
    cpdef RunningWeightedAverage bootstrap(self, size_t n=*, size_t k=*)
    cpdef RunningWeightedAverage subsample(self, size_t k)

    @staticmethod
    cdef RunningWeightedAverage _create(list peaks)
    cdef void _update(self, list peaks)


cdef class RunningWeightedAverageNeutralMass(RunningWeightedAverage):

    @staticmethod
    cdef RunningWeightedAverageNeutralMass _create(list peaks)
